defmodule Nulla.Models.Note do
  use Ecto.Schema
  import Ecto.Changeset
  import Ecto.Query
  alias Nulla.Repo
  alias Nulla.Snowflake
  alias Nulla.Models.Actor
  alias Nulla.Models.MediaAttachment

  @primary_key {:id, :integer, autogenerate: false}
  schema "notes" do
    field :inReplyTo, :string
    field :published, :utc_datetime
    field :url, :string
    field :visibility, :string
    field :to, {:array, :string}
    field :cc, {:array, :string}
    field :sensitive, :boolean, default: false
    field :content, :string
    field :language, :string

    belongs_to :actor, Actor
    has_many :media_attachments, MediaAttachment

    timestamps(type: :utc_datetime)
  end

  @doc false
  def changeset(note, attrs) do
    note
    |> cast(attrs, [
      :id,
      :inReplyTo,
      :published,
      :url,
      :visibility,
      :to,
      :cc,
      :sensitive,
      :content,
      :language,
      :actor_id
    ])
    |> validate_required([:published, :url, :to, :cc, :content, :language, :actor_id])
  end

  def create_note(attrs) when is_map(attrs) do
    id = Map.get(attrs, :id, Snowflake.next_id())
    actor_id = Map.get(attrs, :actor_id)
    actor = Actor.get_actor(id: actor_id)
    published = Map.get(attrs, :published, DateTime.utc_now())
    visibility = Map.get(attrs, :visibility, "public")

    url =
      case Map.get(attrs, :url) do
        nil ->
          "#{actor.url}/#{id}"

        _ ->
          Map.get(attrs, :url)
      end

    {to, cc} =
      case visibility do
        "public" ->
          to = ["https://www.w3.org/ns/activitystreams#Public"]
          cc = [actor.followers | Map.get(attrs, :cc, [])]

          {to, cc}

        "unlisted" ->
          to = [actor.followers]
          cc = ["https://www.w3.org/ns/activitystreams#Public" | Map.get(attrs, :cc, [])]

          {to, cc}

        "followers" ->
          to = [actor.followers]
          cc = Map.get(attrs, :cc, [])

          {to, cc}

        "private" ->
          to = Map.get(attrs, :to, [])
          cc = []

          {to, cc}

        _ ->
          raise ArgumentError, "Invalid visibility: #{visibility}"
      end

    attrs =
      attrs
      |> Map.put(:id, id)
      |> Map.put(:published, published)
      |> Map.put(:url, url)
      |> Map.put(:to, to)
      |> Map.put(:cc, cc)

    %__MODULE__{}
    |> changeset(attrs)
    |> Repo.insert()
  end

  def get_note(by) when is_map(by) or is_list(by) do
    Repo.get_by(__MODULE__, by)
  end

  def get_latest_notes(actor_id, limit \\ 20) do
    from(n in __MODULE__,
      where: n.actor_id == ^actor_id,
      order_by: [desc: n.inserted_at],
      limit: ^limit,
      preload: [:actor, :media_attachments]
    )
    |> Repo.all()
  end

  def get_before_notes(actor_id, max_id, limit \\ 20) do
    from(n in __MODULE__,
      where: n.actor_id == ^actor_id and n.id < ^max_id,
      order_by: [desc: n.inserted_at],
      limit: ^limit,
      preload: [:actor, :media_attachments]
    )
    |> Repo.all()
  end

  def get_total_notes_count(actor_id) do
    from(n in __MODULE__, where: n.actor_id == ^actor_id)
    |> Repo.aggregate(:count, :id)
  end
end
